@@ -15,10 +15,16 @@ from dataclasses import field
 from miniogl.AttachmentLocation import AttachmentLocation
 from miniogl.ControlPoint import ControlPoint
 from miniogl.SelectAnchorPoint import SelectAnchorPoint
+from ogl.OglAggregation import OglAggregation
+from ogl.OglAssociation import OglAssociation
 
 from ogl.OglClass import OglClass
+from ogl.OglComposition import OglComposition
+from ogl.OglInheritance import OglInheritance
+from ogl.OglInterface import OglInterface
 from ogl.OglInterface2 import OglInterface2
 from ogl.OglLink import OglLink
+from ogl.OglNoteLink import OglNoteLink
 
 from pyutmodel.PyutClass import PyutClass
 from pyutmodel.PyutDisplayParameters import PyutDisplayParameters
@@ -147,10 +153,11 @@ class UnTangler:
 
         documentInformation.documentType = pyutDocument['type']
         documentInformation.documentTitle = documentTitle
-        documentInformation.scrollPositionX = pyutDocument['scrollPositionX']
-        documentInformation.scrollPositionY = pyutDocument['scrollPositionY']
-        documentInformation.pixelsPerUnitX = pyutDocument['pixelsPerUnitX']
-        documentInformation.pixelsPerUnitY = pyutDocument['pixelsPerUnitY']
+
+        documentInformation.scrollPositionX = int(pyutDocument['scrollPositionX'])
+        documentInformation.scrollPositionY = int(pyutDocument['scrollPositionY'])
+        documentInformation.pixelsPerUnitX  = int(pyutDocument['pixelsPerUnitX'])
+        documentInformation.pixelsPerUnitY  = int(pyutDocument['pixelsPerUnitY'])
 
         self.logger.debug(f'{documentInformation=}')
 
@@ -262,6 +269,19 @@ class UnTangler:
         return sourceCode
 
     def _graphicLinkToOglLink(self, graphicLink: Element, oglClassDictionary: OglClassDictionary) -> OglLink:
+        """
+        This code is way too convoluted.  Failing to do any of these step in this code leads to BAD
+        visual representations.
+        TODO:  Figure out how to simplify this code and/or make it more readable and obvious on how to create
+        links (of whatever kind) between 2 OglClass'es
+
+        Args:
+            graphicLink:        The XML `GraphicClass` element
+            oglClassDictionary: A dictionary indexed by an ID that returns an appropriate OglClass instance
+
+        Returns:  A fully formed OglLink including control points
+
+        """
 
         assert len(oglClassDictionary) != 0, 'Developer forgot to create dictionary'
         srcX: int = int(graphicLink['srcX'])
@@ -277,19 +297,34 @@ class UnTangler:
         singleLink:  Element = links[0]
         sourceId:    int = int(singleLink['sourceId'])
         dstId:       int = int(singleLink['destId'])
-        self.logger.debug(f'graphicLink= {srcX=} {srcY=} {dstX=} {dstY=} {spline=}')
-
-        srcShape = oglClassDictionary[sourceId]
-        dstShape = oglClassDictionary[dstId]
-        assert srcShape is not None, 'Missing source shape, invalid XML'
-        assert dstShape is not None, 'Missing destination shape, invalid XML'
+        self.logger.debug(f'graphicLink= {srcX=} {srcY=} {dstX=} {dstY=}')
+        try:
+            srcShape = oglClassDictionary[sourceId]
+            dstShape = oglClassDictionary[dstId]
+        except KeyError as ke:
+            self.logger.error(f'Developer Error -- srcId: {sourceId} - dstId: {dstId}  error: {ke}')
+            return cast(OglLink, None)
 
         pyutLink: PyutLink = self._linkToPyutLink(singleLink, source=srcShape.pyutObject, destination=dstShape.pyutObject)
-        oglLink: OglLink = OglLink(srcShape=srcShape, pyutLink=pyutLink, dstShape=dstShape, srcPos=(srcX, srcY), dstPos=(dstX, dstY))
-
+        oglLink:  OglLink  = self._createOglLink(srcShape=srcShape, pyutLink=pyutLink, destShape=dstShape,
+                                                 linkType=pyutLink.linkType,
+                                                 srcPos=(srcX, srcY),
+                                                 dstPos=(dstX, dstY)
+                                                 )
+        srcShape.addLink(oglLink)
+        dstShape.addLink(oglLink)
+        oglLink.SetSpline(spline)
         controlPoints: UntangledControlPoints = self._generateControlPoints(graphicLink=graphicLink)
 
-        parent = oglLink.GetSource().GetParent()
+        # put the anchors at the right position
+        srcAnchor = oglLink.GetSource()
+        dstAnchor = oglLink.GetDestination()
+        srcAnchor.SetPosition(srcX, srcY)
+        dstAnchor.SetPosition(dstX, dstY)
+
+        # add the control points to the line
+        line   = srcAnchor.GetLines()[0]     # only 1 line per anchor in Pyut
+        parent = line.GetSource().GetParent()
         selfLink: bool = parent is oglLink.GetDestination().GetParent()
 
         for controlPoint in controlPoints:
@@ -427,6 +462,45 @@ class UnTangler:
 
         return oglClassDictionary
 
+    def _createOglLink(self, srcShape, pyutLink, destShape, linkType: PyutLinkType, srcPos=None, dstPos=None):
+        """
+        Used to get a OglLink of the given linkType.
+
+        Args:
+            srcShape:   Source shape
+            pyutLink:   Conceptual links associated with the graphical links.
+            destShape:  Destination shape
+            linkType:   The linkType of the link (OGL_INHERITANCE, ...)
+            srcPos:     source position
+            dstPos:     destination position
+
+        Returns:  The requested link
+        """
+        if linkType == PyutLinkType.AGGREGATION:
+            return OglAggregation(srcShape, pyutLink, destShape, srcPos=srcPos, dstPos=dstPos)
+
+        elif linkType == PyutLinkType.COMPOSITION:
+            return OglComposition(srcShape, pyutLink, destShape, srcPos=srcPos, dstPos=dstPos)
+
+        elif linkType == PyutLinkType.INHERITANCE:
+            return OglInheritance(srcShape, pyutLink, destShape)
+
+        elif linkType == PyutLinkType.ASSOCIATION:
+            return OglAssociation(srcShape, pyutLink, destShape, srcPos=srcPos, dstPos=dstPos)
+
+        elif linkType == PyutLinkType.INTERFACE:
+            return OglInterface(srcShape, pyutLink, destShape, srcPos=srcPos, dstPos=dstPos)
+
+        elif linkType == PyutLinkType.NOTELINK:
+            return OglNoteLink(srcShape, pyutLink, destShape)
+
+        elif linkType == PyutLinkType.SD_MESSAGE:
+            assert False, 'Sequence Diagram Messages not supported'
+            # return OglSDMessage(srcShape=srcShape, pyutSDMessage=pyutLink, dstShape=destShape)
+        else:
+            self.logger.error(f"Unknown OglLinkType: {linkType}")
+            return None
+
     def _str2bool(self, strValue: str) -> bool:
         return strValue.lower() in ("yes", "true", "t", "1", 'True')
 
